#include "gideros.h"
#include "lua.h"
#include "lauxlib.h"

#import <AVFoundation/AVSpeechSynthesis.h>

// some Lua helper functions
#ifndef abs_index
#define abs_index(L, i) ((i) > 0 || (i) <= LUA_REGISTRYINDEX ? (i) : lua_gettop(L) + (i) + 1)
#endif

static void luaL_newweaktable(lua_State *L, const char *mode)
{
	lua_newtable(L);			// create table for instance list
	lua_pushstring(L, mode);
	lua_setfield(L, -2, "__mode");	  // set as weak-value table
	lua_pushvalue(L, -1);             // duplicate table
	lua_setmetatable(L, -2);          // set itself as metatable
}

static void luaL_rawgetptr(lua_State *L, int idx, void *ptr)
{
	idx = abs_index(L, idx);
	lua_pushlightuserdata(L, ptr);
	lua_rawget(L, idx);
}

static void luaL_rawsetptr(lua_State *L, int idx, void *ptr)
{
	idx = abs_index(L, idx);
	lua_pushlightuserdata(L, ptr);
	lua_insert(L, -2);
	lua_rawset(L, idx);
}


/*
 we have two tables:

 1.strong (table -> boolean)
 speak -> add, hide - > remove
 to disable GC for visible ttss

 2. weak (ptr -> table)
 to access table from pointer
 
 */

static char keyStrong = ' ';
static char keyWeak = ' ';

static const char *TTS_UTTERANCE_COMPLETE = "ttsUtteranceComplete";
static const char *TTS_UTTERANCE_STARTED = "ttsUtteranceStarted";
static const char *TTS_ERROR = "ttsError";
static const char *TTS_INIT_COMPLETE = "ttsInitComplete";



/* 
 require "tts"
 

 local tts = TTS.new("en-us") OR TTS.new({language="en-us", speed=1, pitch=1})  OR TTS.new("en-us",1,1) OR TTS.new()
 tts:speak("utterance",utteranceId)
 
 --************** Additional methods *******
 tts:stop()
 tts:setSpeed(1) --Lua input changed to 1 to be consistent with Android. ios native is 0.5
 tts:setPitch(1)
 tts:setVolume(1)
 tts:setLanguage("en-gb")
 tts:setVoiceIdentifier("com.apple.ttsbundle.siri_female_en-GB_premium")
 
 tts:getSpeed()
 tts:getPitch()
 tts:getVolume()
 tts:getVoicesInstalled() - returns  array of key values for each voice: name, identifier, quality, language
 
 tts:addEventListener(Event.TTS_UTTERANCE_COMPLETE, ...)
 tts:addEventListener(Event.TTS_UTTERANCE_STARTED, ...)
 
*/

class TTS;


@interface TTSSpeechUtterance : AVSpeechUtterance
@property (nonatomic, retain ) NSString * utteranceId;
@end

@implementation TTSSpeechUtterance
@end

@interface TTSSpeechSynthesizer : AVSpeechSynthesizer
@property (nonatomic, retain ) NSString * language;
@property (nonatomic, retain ) NSString * voiceIdentifier;
@property (nonatomic, retain ) AVSpeechSynthesisVoice * voice;
@property (nonatomic) float  pitch;
@property (nonatomic) float  rate;
@property (nonatomic) float  volume;
@property (nonatomic) bool   stopped;
@end

@implementation TTSSpeechSynthesizer
@end


@interface TTSDelegate : NSObject<AVSpeechSynthesizerDelegate>
{
}

- (id)initWithTTS:(TTS *)tts;

@property (nonatomic, assign) TTS *tts;
@end

class TTS : public GEventDispatcherProxy
{
    
public:
    
	TTS(lua_State *L) : L(L)
	{
        synthesizer_ = [[TTSSpeechSynthesizer alloc] init];
        synthesizer_.voiceIdentifier=@"com.apple.ttsbundle.siri_female_en-US_compact" ;
        synthesizer_.stopped=false;
        if (lua_istable(L,1))
        {
            lua_getfield(L,1,"language");
            lua_getfield(L,1,"speed");
            lua_getfield(L,1,"pitch");
            synthesizer_.language= [NSString stringWithUTF8String:luaL_optstring(L,-3,"en-US")];

            synthesizer_.rate=luaL_optnumber(L,-2,1.0)*0.5;
            synthesizer_.pitch=luaL_optnumber(L,-1,1.0);
            lua_pop(L,3);
        }
        else
        {
            synthesizer_.language= [NSString stringWithUTF8String:luaL_optstring(L,1,"en-US")];
            synthesizer_.rate=luaL_optnumber(L,2,1.0)*0.5;
            synthesizer_.pitch=luaL_optnumber(L,3,1.0);
        }
 		
        synthesizer_.voice = [AVSpeechSynthesisVoice voiceWithLanguage:synthesizer_.language];
		delegate_ = [[TTSDelegate alloc] initWithTTS:this];
		synthesizer_.delegate = delegate_;

	}

    
    void shutdown()
    {
        synthesizer_.stopped=true;
        [synthesizer_ stopSpeakingAtBoundary:AVSpeechBoundaryImmediate];

    }

	virtual ~TTS()
	{
        NSLog(@"~TTS()");
        shutdown();
		synthesizer_.delegate = nil;
        delegate_.tts = NULL;
        [synthesizer_ release];
		[delegate_ release];
	}
	
	void speak(lua_State *L)
	{
        synthesizer_.stopped=false;
        
        TTSSpeechUtterance *utterance = [[TTSSpeechUtterance alloc] initWithString:[NSString stringWithUTF8String:luaL_optstring(L,2,"")]];
        utterance.utteranceId= [NSString stringWithUTF8String:luaL_optstring(L,3,"")];
        [utterance setRate:synthesizer_.rate];
        [utterance setPitchMultiplier:synthesizer_.pitch];
        [utterance setVolume:synthesizer_.volume];
        utterance.voice = synthesizer_.voice;
        
        [synthesizer_ speakUtterance:utterance];
		
	}
    

    
    void stop(lua_State *L)
    {
        synthesizer_.stopped=true;
        [synthesizer_ stopSpeakingAtBoundary:AVSpeechBoundaryImmediate];
    }
	
    void setLanguage(lua_State *L)
    {
        synthesizer_.language = [NSString stringWithUTF8String:luaL_checkstring(L, -1)];
        synthesizer_.voice=[AVSpeechSynthesisVoice voiceWithLanguage:synthesizer_.language];
    }
    
    void setVoiceIdentifier(lua_State *L)
    {
        synthesizer_.voiceIdentifier = [NSString stringWithUTF8String:luaL_checkstring(L, -1)];
        synthesizer_.voice=[AVSpeechSynthesisVoice voiceWithIdentifier:synthesizer_.voiceIdentifier];
    }
    void setSpeed(lua_State *L)
    {
        float rateReceived=lua_tonumber(L, -1);
        synthesizer_.rate = MIN(MAX(rateReceived*0.5,AVSpeechUtteranceMinimumSpeechRate),AVSpeechUtteranceMaximumSpeechRate);
    }
    void setPitch(lua_State *L)
    {
        synthesizer_.pitch = MIN(MAX((float) lua_tonumber(L, -1),0.5),2);
    }
    
    void setVolume(lua_State *L)
    {
        synthesizer_.volume = MIN(MAX((float) lua_tonumber(L, -1),0),1);
    }
    
    float getSpeed(lua_State *L)
    {
        return synthesizer_.rate*2;
    }
    
    float getPitch(lua_State *L)
    {
        return synthesizer_.pitch;
    }
    
    float getVolume(lua_State *L)
    {
        return synthesizer_.volume;
    }
    
	void utteranceComplete(NSString * utteranceId)
    {   if (!synthesizer_.stopped){
        NSString *state = @"done";
		dispatchEvent(TTS_UTTERANCE_COMPLETE, utteranceId, state);
        }
	}
    
    void utteranceStarted(NSString * utteranceId)
    {
        if (!synthesizer_.stopped){
        NSString *state = @"start";
        dispatchEvent(TTS_UTTERANCE_COMPLETE, utteranceId, state);
      //  dispatchEvent(TTS_UTTERANCE_STARTED, utteranceId, state);
        }
    }
	
	void dispatchEvent(const char *type, NSString *utteranceId, NSString *state)
	{
        
        luaL_rawgetptr(L, LUA_REGISTRYINDEX, &keyWeak);
        luaL_rawgetptr(L, -1, this);
		
        if (lua_isnil(L, -1))
        {
            lua_pop(L, 2);
            return;
        }

        lua_getfield(L, -1, "dispatchEvent");
		
        lua_pushvalue(L, -2);

        lua_getglobal(L, "Event");
        lua_getfield(L, -1, "new");
        lua_remove(L, -2);
	
        lua_pushstring(L, type);
        lua_call(L, 1, 1);
		
		if (utteranceId)
		{
            //lua_pushstring(L, [utteranceId UTF8String]);
            lua_pushstring(L, [utteranceId UTF8String]);

            lua_setfield(L, -2, "utteranceId");
		}
		
        if (state)
        {
            //any parameters are pushed into the table
            //returned from "new"
            //this table will be sent to the Lua Event function
            lua_pushstring(L, [state UTF8String]);
            lua_setfield(L, -2, "state");
        }
		
		if (lua_pcall(L, 2, 0, 0) != 0)
			g_error(L, lua_tostring(L, -1));
		
		lua_pop(L, 2);
	}
	
private:
	lua_State *L;
	
	TTSDelegate *delegate_;
	TTSSpeechSynthesizer *synthesizer_;
	
};

@implementation TTSDelegate

@synthesize tts = tts_;


- (id)initWithTTS:(TTS *)tts
{
	if (self = [super init])
	{
        tts_ = tts;
	}
	
	return self;
}

- (void)dealloc
{
   // tts_->shutdown();
    [AVSpeechUtterance release];
    [TTSSpeechSynthesizer release];
    [TTSDelegate release];
    [super dealloc];
}

-(void)speechSynthesizer:(TTSSpeechSynthesizer *)synthesizer didFinishSpeechUtterance:(TTSSpeechUtterance *)utterance {
    NSLog(@"Playback finished");
    tts_->utteranceComplete(utterance.utteranceId);
}

-(void)speechSynthesizer:(TTSSpeechSynthesizer *)synthesizer didStartSpeechUtterance:(TTSSpeechUtterance *)utterance {
    NSLog(@"Playback started");
    tts_->utteranceStarted(utterance.utteranceId);
}

@end

static int create(lua_State *L)
{
		
	TTS *tts = new TTS(L);
	g_pushInstance(L, "TTS", tts->object());
	
	luaL_rawgetptr(L, LUA_REGISTRYINDEX, &keyWeak);
	lua_pushvalue(L, -2);
	luaL_rawsetptr(L, -2, tts);
	lua_pop(L, 1);
	
	return 1;
}

static int destruct(lua_State *L)
{
	void *ptr = *(void**)lua_touserdata(L, 1);
	GReferenced *object = static_cast<GReferenced*>(ptr);
	TTS *tts = static_cast<TTS*>(object->proxy());
	
	tts->unref();
	
	return 0;
}

static TTS *getTTSInstance(lua_State *L, int index)
{
	GReferenced *object = static_cast<GReferenced*>(g_getInstance(L, "TTS", index));
	TTS *tts = static_cast<TTS*>(object->proxy());
	
	return tts;
}



static int speak(lua_State *L)
{
    TTS *tts = getTTSInstance(L, 1);
    tts->speak(L);


	return 0;
}

static int stop(lua_State*L)
{
    TTS *tts = getTTSInstance(L, 1);
    tts->stop(L);
    return 0;
}

static int setLanguage(lua_State *L)
{
    TTS *tts = getTTSInstance(L, 1);
    tts->setLanguage(L);
    return 0;
}

static int setVoiceIdentifier(lua_State *L)
{
    TTS *tts = getTTSInstance(L, 1);
    tts->setVoiceIdentifier(L);
    return 0;
}


static int getVoicesInstalled(lua_State*L)
{
    ///return table array with keys "name", "identifier" (to setVoiceName), quality, and locale
    NSArray *allVoices = [AVSpeechSynthesisVoice speechVoices];
    int arrLength= (int) allVoices.count;
    lua_createtable(L, arrLength, 0);
    
    int index = 1;
    for (AVSpeechSynthesisVoice *voice in allVoices) {
         // NSLog(@"Voice Name: %@, Identifier: %@, Quality: %ld@, Language: %@", voice.name, voice.identifier, (long)voice.quality, voice.language);
        lua_pushnumber(L, index++);
        lua_createtable(L, 0, 2);
        lua_pushstring(L, [voice.identifier UTF8String]);
        lua_setfield(L, -2, "identifier");
        lua_pushstring(L, [voice.name UTF8String]);
        lua_setfield(L, -2, "name");
        lua_pushnumber(L, voice.quality);
        lua_setfield(L, -2, "quality");
        lua_pushstring(L, [voice.language UTF8String]);
        lua_setfield(L, -2, "language");
        lua_settable(L, -3);
    }
    return 1;
}

static int setPitch(lua_State *L)
{
    TTS *tts = getTTSInstance(L, 1);
    tts->setPitch(L);
    return 0;
}

static int getPitch(lua_State *L)
{
    TTS *tts = getTTSInstance(L, 1);
    float pitchMultiplier = tts->getPitch(L);
    lua_pushnumber(L, pitchMultiplier);
    return 1;
}

static int setSpeed(lua_State *L)
{
    TTS *tts = getTTSInstance(L, 1);
    tts->setSpeed(L);
    return 0;
}

static int getSpeed(lua_State *L)
{
    TTS *tts = getTTSInstance(L, 1);
    float speed = tts->getSpeed(L);
    lua_pushnumber(L, speed);
    return 1;
}

static int setVolume(lua_State *L)
{
    TTS *tts = getTTSInstance(L, 1);
    tts->setVolume(L);
    return 0;
}

static int getVolume(lua_State *L)
{
    TTS *tts = getTTSInstance(L, 1);
    float volume = tts->getVolume(L);
    lua_pushnumber(L, volume);
    return 1;
}

static int shutdown(lua_State *L)
{
    ////*********Currently only stops playback
    TTS *tts = getTTSInstance(L, 1);
    tts->stop(L);
    return 0;
}

static int loader(lua_State *L)
{
	const luaL_Reg functionlist[] = {
		{"speak", speak},
        {"stop", stop},
        {"setLanguage", setLanguage},
        {"setVoiceIdentifier", setVoiceIdentifier},
        {"setPitch", setPitch},
        {"setSpeed", setSpeed},
        {"setVolume", setVolume},
        {"getPitch", getPitch},
        {"getSpeed", getSpeed},
        {"getVolume", getVolume},
        {"getVoicesInstalled", getVoicesInstalled},
        {"shutdown", shutdown},
		{NULL, NULL},
	};
	
	g_createClass(L, "TTS", "EventDispatcher", create, destruct, functionlist);
    
    lua_newtable(L);
    luaL_rawsetptr(L, LUA_REGISTRYINDEX, &keyStrong);

    luaL_newweaktable(L, "v");
    luaL_rawsetptr(L, LUA_REGISTRYINDEX, &keyWeak);

	lua_getglobal(L, "Event");
	lua_pushstring(L, TTS_UTTERANCE_COMPLETE);
	lua_setfield(L, -2, "TTS_UTTERANCE_COMPLETE");
	lua_pushstring(L, TTS_UTTERANCE_STARTED);
	lua_setfield(L, -2, "TTS_UTTERANCE_STARTED");
    lua_pushstring(L, TTS_ERROR);
    lua_setfield(L, -2, "TTS_ERROR");
    lua_pushstring(L, TTS_INIT_COMPLETE);
    lua_setfield(L, -2, "TTS_INIT_COMPLETE");
	lua_pop(L, 1);
		
	lua_getglobal(L, "tts");

	return 1;
}

static void g_initializePlugin(lua_State* L)
{
	lua_getglobal(L, "package");
	lua_getfield(L, -1, "preload");
	
	lua_pushcfunction(L, loader);
	lua_setfield(L, -2, "tts");
	
	lua_pop(L, 2);	
}

static void g_deinitializePlugin(lua_State *L)
{
}

REGISTER_PLUGIN("TTS", "1.1")
